;compiler: avrasm
	.include "m32def.inc"
	
	.cseg
	.org	0x0000
	rjmp	Demo	;jump to demo code
	.org	0x000E
	rjmp	INTmeasurePeriod	;demo counter measure period
	.org	0x0016
	rjmp	INTtmr0ovf	;demo counter extension
	
;************************************************************************************************************************
; HD44780 LCD routines 4-wires interface, no RW pin used - constant delay 40us
; don't forget set correct XTAL freq somewhere
; don't forget init ddrx registers for E,RS and DATA pins immediately after MCU reset!!!
;
; author: Stanislav Maslan, s.maslan@seznam.cz 
;************************************************************************************************************************
	.include "macros.asm"
	.equ	XTAL=16000000	;clock freq in Hz
	
	;pin/port definitions
	.equ	LCDportE=portd	;E pin port
	.equ	LCK=5	;pin?
	.equ	LCDportRS=portd	;RS pin port
	.equ	LRS=6	;pin?
	.equ	LCDportDATA=portc	;DATA pins port
	.equ	LCDportDATAnib=0	;low=0/high=1 nibble of the port

;------------------------------------------------------------------------------------------------------------------------
;	LCD initialization, call this routine first but generate at least 20ms delay from power on!
;	R16=1 one line, R16=2 two lines
;------------------------------------------------------------------------------------------------------------------------
LCDinit:	cbi	LCDportRS,LRS	;instructions
	push	R16
	;
	in	R0,sreg
	cli		;disable IRQ
	;back to 8-bit, if it was reset (brown out detector or ISP reset)
	ldi	R16,0b00000011	;to 8-bit mode
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	ldi	R16,0b00000000	;if it was 4-bit mode (after reset) complete instruction with low nibble 0b0000
	rcall	LCDnibOut	;if it was 8-bit mode, 0b00000000 instruction will be ignored
	rcall	LCDclock	;clock
	rcall	LCDwait164	;wait 1,64ms
	;now we are surely in 8-bit mode, init 4-bit mode
	ldi	R16,0b00000010
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	;
	out	sreg,R0
	;now select mode (R16=1 one line, R16=2 two lines)
	pop	R16
	dec	R16	;N--
	swap	R16
	lsr	R16	;N<<3
	andi	R16,(1<<3)
	ori	R16,0b00100000	;add init mask
	rcall	LCDwrInstr	;full init again into 4-bit mode, selected number of lines
	ldi	R16,0b00001100
	rcall	LCDwrInstr	;turn on display
	ldi	R16,0b00000110
	rcall	LCDwrInstr	;set cursor shift mode
	rjmp	LCDclrAll	;clear all
;------------------------------------------------------------------------------------------------------------------------
;	LCD write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines
;------------------------------------------------------------------------------------------------------------------------
	;with white leading zeroes
LCDwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	LCDwrBCD16
	;with leading zeroes
LCDwrBCD16z:	set
	rjmp	LCDwrBCD16
	;skip leading zeroes
LCDwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
LCDwrBCD16:	mov	R16,R15
	rcall	LCDwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	LCDwrCharBCD
	mov	R16,R14
	rcall	LCDwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	LCDwrCharBCD
	set
	mov	R16,R13
	rcall	LCDwrCharBCD
	ret
	;common write BCD num.
LCDwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	LCDwrCharBCDnz	;zero?
	brts	LCDwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	LCDwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	LCDwrChar	;write number
LCDwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	LCDwrChar	;write number
LCDwrCharBCDzc:	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Z pointed (in bytes!) 0xFF terminated string from FLASH, 0x00 line break
;	.db "text for 1st line",0x00,"text for second line",0xFF 
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrFlash:	lpm	R16,Z+	;get char
	rcall	LCDwrChar
	lpm	R16,Z
	cpi	R16,0x00
	brne	LCDwrStrFlashNBL
	adiw	R30,1
	ldi	R16,0x40
	rcall	LCDsetPos
	rjmp	LCDwrStrFlash
LCDwrStrFlashNBL:	cpi	R16,0xFF	;str end?
	brne	LCDwrStrFlash	;next char
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Y pointed RAM array, R24=length
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrRAM:	ld	R16,Y+
	rcall	LCDwrChar
	dbnz	R24,LCDwrStrRAM
	ret
;------------------------------------------------------------------------------------------------------------------------
;	set pos (R16):
;	1.line from 0x00
;	2.line from 0x40
;	3.line from 0x14
;	4.line from 0x54
;------------------------------------------------------------------------------------------------------------------------
LCDsetPos:	ori	R16,1<<7
	rjmp	LCDwrInstr
;------------------------------------------------------------------------------------------------------------------------
;	clr all
;------------------------------------------------------------------------------------------------------------------------
LCDclrAll:	ldi	R16,0b00000001
	rcall	LCDwrInstr
	rjmp	LCDwait164
;------------------------------------------------------------------------------------------------------------------------
;	clr R16 line, set cursor left on the line
;------------------------------------------------------------------------------------------------------------------------
LCDclrLine:	cpi	R16,1
	brne	LCDclrLineNo1
	ldi	R24,40	;40 chars
	ldi	R16,0x00	;1.line
	rjmp	LCDclrLineLC
LCDclrLineNo1:	cpi	R16,2
	brne	LCDclrLineNo2
	ldi	R24,40	;40 chars
	ldi	R16,0x40	;2.line
	rjmp	LCDclrLineLC
LCDclrLineNo2:	cpi	R16,3
	brne	LCDclrLineNo3
	ldi	R24,20	;20 chars
	ldi	R16,0x14	;3.line
	rjmp	LCDclrLineLC		
LCDclrLineNo3:	ldi	R24,20	;20 chars
	ldi	R16,0x54	;4.line
	rjmp	LCDclrLineLC
	;common clr routine
LCDclrLineLC:	push	R16
	rcall	LCDsetPos
LCDclrLineL:	ldi	R16,' '
	rcall	LCDwrChar
	dbnz	R24,LCDclrLineL
	pop	R16
	rcall	LCDsetPos
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 40us
;------------------------------------------------------------------------------------------------------------------------	
LCDwait40:	ldi	R16,XTAL/74000
LCDwait40_1:	dbnz	R16,LCDwait40_1
	ret
;------------------------------------------------------------------------------------------------------------------------
;	wait 1.64ms
;------------------------------------------------------------------------------------------------------------------------	
LCDwait164:	ldi	R17,41	;41*40us
LCDwait164_2:	ldi	R16,XTAL/74000
LCDwait164_1:	dbnz	R16,LCDwait164_1
	dbnz	R17,LCDwait164_2
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write R16 char, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrChar:	sbi	LCDportRS,LRS	;data
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 instr, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrInstr:	cbi	LCDportRS,LRS	;instructions
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 byte
;------------------------------------------------------------------------------------------------------------------------
LCDwrByte:	in	R0,sreg
	cli		;disable IRQ
	;high nibble
	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	out	sreg,R0
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write low nibble R16 to the LCD data port
;------------------------------------------------------------------------------------------------------------------------
LCDnibOut:	andi	R16,0b00001111	;mask low nibble
	.if	LCDportDATAnib==0	;low nibble?
	;low nibble
	in	R17,LCDportDATA
	andi	R17,0b11110000
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.elif	LCDportDATAnib==1	;high nibble?
	;high nibble
	swap	R16
	in	R17,LCDportDATA
	andi	R17,0b00001111
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.else
	.error "low=0 or high=1 nibble for LCD???"
	.endif
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD clock
;------------------------------------------------------------------------------------------------------------------------	
LCDclock:	sbi	LCDportE,LCK
	nop
	nop
	cbi	LCDportE,LCK
	ret
	
	
	
	
	





;************************************************************************************************************************
;************************************************************************************************************************
;	democode
;************************************************************************************************************************
;************************************************************************************************************************		
	.dseg
FreqTmp:	.byte	1
Freq:	.byte	2


	.cseg
;************************************************************************************************************************
;	demo counter extension INT
;************************************************************************************************************************
INTtmr0ovf:	push	R16
	in	R16,sreg
	push	R16
	;
	lds	R16,FreqTmp
	inc	R16
	sts	FreqTmp,R16
	;
	pop	R16
	out	sreg,R16
	pop	R16
	reti	
;************************************************************************************************************************
;	demo coutner measure period INT
;************************************************************************************************************************	
INTmeasurePeriod:	pushw	R16
	in	R16,sreg
	push	R16
	;read value
	lds	R17,FreqTmp
	outi	tccr0,0	;stop counting
	in	R16,gifr
	sbrs	R16,tov0
	rjmp	PC+2
	inc	R17
	in	R16,tcnt0
	stsw	Freq,R16
	;restart counting
	outi	tcnt0,0
	outi	gifr,(1<<tov0)
	stsi	FreqTmp,0
	outi	sfior,psr10
	ldiw	R16,0
	outw	tcnt1,R16
	outi	tccr0,0b00000111
	;
	pop	R16
	out	sreg,R16
	popw	R16
	reti
	
	
;------------------------------------------------------------------------------------------------------------------------
;	prog init
;------------------------------------------------------------------------------------------------------------------------	
Demo:	outi	ddrc,0b00001111	;LCD data pins
	outi	ddrd,(1<<LRS)|(1<<LCK)	;LCD ctrl pins
	
	cbi	portd,LCK	;LCD E pin = 0 (don't forget this!)
	
	ldiw	R16,ramend
	outw	sp,R16	;set stack
	
	
	ldi	R16,2
	rcall	DelayVar	;wait 20ms
		
	ldi	R16,2
	rcall	LCDinit	;init LCD as 2 line
	
	ldiw	R30,Txt1T*2
	rcall	LCDwrStrFLASH
	
	ldi	R16,200
	rcall	DelayVar	;wait 2s
		
	
	ldi	R16,1
	rcall	LCDclrLine	;clear 1.line
	ldiw	R30,Txt3T*2
	rcall	LCDwrStrFLASH	
	
	
	ldiw	R16,583
	rcall	bin2BCD16p5	;number to BCD
	
	ldi	R16,2
	rcall	LCDclrLine	;clear 2.line
	rcall	LCDwrBCD16z	;write num, with leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	ldi	R16,2
	rcall	LCDclrLine	;clear 2.line
	rcall	LCDwrBCD16wz	;write num, white leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	ldi	R16,2
	rcall	LCDclrLine	;clear 2.line
	rcall	LCDwrBCD16sz	;write num, skip leading zeroes
	ldi	R16,200
	rcall	DelayVar	;wait 2s
	
	rcall	LCDclrAll	;clr all
	
	
	;now one debug routine for easy FLASH string writing
	rcall	DwrStrFLASH
	.db	"This'll be skipped",0x00,"by routine ...",0xFF
	;here you can place your next code - .db will be skipped by routine "DwrStrFLASH"	
	ldi	R16,200
	rcall	DelayVar	;wait
			
	rcall	LCDclrAll	;clr all
	ldiw	R30,Txt2T*2
	rcall	LCDwrStrFLASH	;next string
	
	ldi	R16,200
	rcall	DelayVar	;wait
	
	rcall	LCDclrAll	;clr all
	ldiw	R30,Txt4T*2
	rcall	LCDwrStrFLASH	;next string
	
	ldi	R16,200
	rcall	DelayVar	;wait
	
	;simple counter demo, input at T0 (portb,0) ...	
	outi	tccr1a,0
	outi	tccr1b,0b00001101
	ldiw	R16,XTAL/1024-1
	outw	ocr1a,R16	;timer 1: T=1s
	
	outi	tccr0,0b00000111	;timer 0: external clk
	
	outi	timsk,(1<<ocie1a)|(1<<toie0)	;timer 1 CTC IRQ, timer 0 overflow IRQ
	
	ldiw	R16,0
	stsw	Freq,R16
	stsi	FreqTmp,0
	clrw	R4	;clear top freq
	
	sei
		
	;write loop
Loop1:	ldi	R16,0x40
	rcall	LCDsetPos	;return cursor
	cli
	ldsw	R16,Freq	;get freq
	sei
	cpw	R4,R16
	brsh	LoopS1
	movw	R4,R16
LoopS1:	rcall	bin2BCD16p5	;to BCD
	rcall	LCDwrBCD16wz	;write
	ldiw	R30,Txt5T*2
	rcall	LCDwrStrFLASH	;write extension
	
	ldiw	R30,Txt6T*2
	rcall	LCDwrStrFLASH	;(max.
	movw	R16,R4
	rcall	bin2BCD16p5
	rcall	LCDwrBCD16sz	;max freq
	ldiw	R30,Txt7T*2
	rcall	LCDwrStrFLASH	;Hz)
	
	ldi	R16,20
	rcall	DelayVar	;wait 200ms
	rjmp	Loop1	;wait forever
	
	;text strings
Txt1T:	.db	"HD44780 routines by",0x00,"Stanislav Maslan",0xFF
Txt2T:	.db	"String skipped",0x00,"successfully :-)",0xFF
Txt3T:	.db	"BCD routines:",0xFF
Txt4T:	.db	"Simple counter (T0):",0xFF
Txt5T:	.db	"Hz",0xFF
Txt6T:	.db	"(max.",0xFF
Txt7T:	.db	"Hz)",0xFF

;************************************************************************************************************************
;	other
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
;	write FLASH string which is placed directly after rcall instr to this routine
;	(this routine can be used only in devices up to 64kB FLASH!!!)
;------------------------------------------------------------------------------------------------------------------------
DwrStrFLASH:	in	R0,sreg
	cli		;disable IRQ
	;	
	inw	R28,sp	;read stack PTR
	adiw	R28,3
	ldw	R30,-Y	;return address from this routine
	lslw	R30
	pushw	R30	;string address
	;find terminator 0xFF in the string
DwrStrFLASHL1:	lpm	R16,Z+
	cpi	R16,0xFF
	brne	DwrStrFLASHL1
	;Z odd?
	sbrs	R30,0
	rjmp	DwrStrFLASHZnn
	adiw	R30,1	;++ to even number
DwrStrFLASHZnn:	lsrw	R30	;strend to words
	st	Y+,R31
	st	Y+,R30	;write new ret address :-D, now ret instr will jump behind string
	
	popw	R30	;text string address
	rcall	LCDwrStrFLASH	;write string
	;
	out	sreg,R0	
DwrStrFLASHdata:	ret		;return behind string	
	
;************************************************************************************************************************
;	delay routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	variable long delay (R16:R17 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayLongVar:	pushw	R18
DelayLongVar_3:	ldi	R19,250
DelayLongVar_2:	ldi	R18,XTAL/74000
DelayLongVar_1:	dbnz	R18,DelayLongVar_1
	dbnz	R19,DelayLongVar_2
	dbnzw	R16,DelayLongVar_3
	popw	R18
	ret
;------------------------------------------------------------------------------------------------------------------------
;	variable delay (R16 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayVar:	pushw	R18
DelayVar_3:	ldi	R19,250
DelayVar_2:	ldi	R18,XTAL/74000
DelayVar_1:	dbnz	R18,DelayVar_1
	dbnz	R19,DelayVar_2
	dbnz	R16,DelayVar_3
	popw	R18
	ret
;************************************************************************************************************************
;	math routines
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
; bin2BCD16p5 - 16-bit binary to 5-digit packed BCD conversion (0..65535)
; prevod 16-bit binarniho cisla na peticiferne cislo v BCD
;
; BCD16res0:BCD16res1:BCD16res2 <- bin16r0:bin16r1
; BCD(0x01 0x23 0x45) <- bin(12345)
;      ^         ^
; BCD16res2  BCD16res0
;
; words:  29 + ret
; cycles: 29/180 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD16res0=R13
	.def	BCD16res1=R14
	.def	BCD16res2=R15
	.def	bin16r0=R16
	.def	bin16r1=R17
	.def	BB16_cnt1=R30
	.def	BB16_cnt2=R31
	;
bin2BCD16p5:	ser	BB16_cnt2
	mov	BCD16res2,BB16_cnt2
BB16p5_L10k:	inc	BCD16res2
	subi	bin16r0,low(10000)
	sbci	bin16r1,high(10000)
	brcc	BB16p5_L10k
	subi	bin16r0,low(-10000)
	sbci	bin16r1,high(-10000)
	ldi	BB16_cnt1,(256-16)
BB16p5_L1k:	subi	BB16_cnt1,(-16)
	subi	bin16r0,low(1000)
	sbci	bin16r1,high(1000)
	brcc	BB16p5_L1k
	subi	bin16r0,low(-1000)
	sbci	bin16r1,high(-1000)
	mov	BCD16res1,BB16_cnt2
BB16p5_L100:	inc	BCD16res1
	subi	bin16r0,low(100)
	sbci	bin16r1,high(100)
	brcc	BB16p5_L100
	subi	bin16r0,low(-100)
	or	BCD16res1,BB16_cnt1
	ldi	BB16_cnt1,(256-16)
BB16p5_L10:	subi	BB16_cnt1,(-16)
	subi	bin16r0,10
	brcc	BB16p5_L10
	subi	bin16r0,-10
	mov	BCD16res0,BB16_cnt1
	or	BCD16res0,bin16r0
	ret